# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl._core.quaternion import Quaternion
from sisl.shape.ellipsoid import *
from sisl.utils.mathematics import fnorm

pytestmark = pytest.mark.shape


@pytest.mark.filterwarnings("ignore", message="*orthogonalizes the vectors")
def test_create_ellipsoid():
    el = Ellipsoid(1.0)
    el = Ellipsoid([1.0, 1.0, 1.0])
    el = Ellipsoid([1.0, 1.0, 1.0], [1.0] * 3)
    el = Ellipsoid([1.0, 2.0, 3.0])
    v0 = [1.0, 0.2, 1.0]
    v1 = [1.0, -0.2, 1.0]
    v2 = [1.0, -0.2, -1.0]
    el = Ellipsoid([v0, v1, v2])
    q = Quaternion(45, [1, 0, 0], rad=False)
    eye = np.identity(3)
    eye = q.rotate(eye)
    el = Ellipsoid(eye)
    e2 = el.scale(1.1)
    assert np.allclose(el.radius + 0.1, e2.radius)
    e2 = el.scale([1.1] * 3)
    assert np.allclose(el.radius + 0.1, e2.radius)
    e2 = el.scale([1.1, 2.1, 3.1])
    assert np.allclose(el.radius + [0.1, 1.1, 2.1], e2.radius)
    assert el.expand(2).volume == pytest.approx(4 / 3 * np.pi * 3**3)
    str(el)


def test_tosphere():
    el = Ellipsoid([1.0, 1.0, 1.0])
    assert el.to.Sphere().radius == pytest.approx(1)
    el = Ellipsoid([1.0, 2.0, 3.0])
    assert el.to.Sphere().radius == pytest.approx(3)


@pytest.mark.filterwarnings("ignore", message="*orthogonalizes the vectors")
def test_create_ellipsoid_fail():
    v0 = [1.0, 0.2, 1.0]
    v1 = [1.0, 0.2, 1.0]
    v2 = [1.0, -0.2, -1.0]
    with pytest.raises(ValueError):
        el = Ellipsoid([v0, v1, v2])


def test_create_ellipsoid_fail2():
    v0 = [1.0, 0.2, 1.0]
    v1 = [1.0, 0.2, 1.0]
    v2 = [1.0, -0.2, -1.0]
    v3 = [1.0, -0.2, -1.0]
    with pytest.raises(ValueError):
        el = Ellipsoid([v0, v1, v2, v3])


def test_create_sphere():
    el = Sphere(1.0)
    el = Sphere(1.0, center=[1.0] * 3)
    assert el.volume == pytest.approx(4 / 3 * np.pi)
    assert el.scale(2).volume == pytest.approx(4 / 3 * np.pi * 2**3)
    assert el.expand(2).volume == pytest.approx(4 / 3 * np.pi * 3**3)


def test_scale1():
    e1 = Ellipsoid([1.0, 1.0, 1.0])
    e2 = e1.scale(1.1)
    assert np.allclose(e1.radius + 0.1, e2.radius)
    e2 = e1.scale([1.1] * 3)
    assert np.allclose(e1.radius + 0.1, e2.radius)
    e2 = e1.scale([1.1, 2.1, 3.1])
    assert np.allclose(e1.radius + [0.1, 1.1, 2.1], e2.radius)


def test_expand1():
    e1 = Ellipsoid([1.0, 1.0, 1.0])
    e2 = e1.expand(1.1)
    assert np.allclose(e1.radius + 1.1, e2.radius)
    e2 = e1.expand([1.1] * 3)
    assert np.allclose(e1.radius + 1.1, e2.radius)
    e2 = e1.expand([1.1, 2.1, 3.1])
    assert np.allclose(e1.radius + [1.1, 2.1, 3.1], e2.radius)


def test_expand_fail():
    el = Ellipsoid(1)
    with pytest.raises(ValueError):
        el.expand([1, 2])


def test_within1():
    o = Ellipsoid([1.0, 2.0, 3.0])
    assert not o.within([-1.0] * 3).any()
    assert o.within([0.2] * 3)
    assert o.within([0.5] * 3)
    o = Ellipsoid([1.0, 1.0, 2.0])
    assert not o.within([-1.0] * 3).any()
    assert o.within([0.2] * 3)
    assert o.within([0.5] * 3)
    o = Sphere(1.0)
    assert not o.within([-1.0] * 3).any()
    assert o.within([0.2] * 3)
    assert o.within([0.5] * 3)


def test_within_index1():
    o = Ellipsoid([1.0, 2.0, 3.0])
    assert o.within_index([-1.0] * 3).size == 0
    assert o.within_index([0.2] * 3) == [0]
    assert o.within_index([0.5] * 3) == [0]
    o = Ellipsoid([1.0, 1.0, 2.0])
    assert o.within_index([-1.0] * 3).size == 0
    assert o.within_index([0.2] * 3) == [0]
    assert o.within_index([0.5] * 3) == [0]
    o = Sphere(1.0)
    assert o.within_index([-1.0] * 3).size == 0
    assert o.within_index([0.2] * 3) == [0]
    assert o.within_index([0.5] * 3) == [0]


def test_sphere_and():
    # For all intersections after
    v = np.array([1.0] * 3)
    v /= fnorm(v)
    D = np.linspace(0, 5, 50)
    inside = np.ones(len(D), dtype=bool)

    A = Sphere(2.0)
    is_first = True
    inside[:] = True
    for i, d in enumerate(D):
        B = Sphere(1.0, center=v * d)
        C = (A & B).to.Sphere()
        if is_first and C.radius < B.radius:
            is_first = False
            inside[i:] = False
        if inside[i]:
            assert C.radius == pytest.approx(B.radius)
        else:
            assert C.radius < B.radius

    A = Sphere(0.5)
    is_first = True
    inside[:] = True
    for i, d in enumerate(D):
        B = Sphere(1.0, center=v * d)
        C = (A & B).to.Sphere()
        str(A) + str(B) + str(C)
        if is_first and C.radius < A.radius:
            inside[i:] = False
            is_first = False
        if inside[i]:
            assert C.radius == pytest.approx(A.radius)
        else:
            assert C.radius < A.radius


def test_translate():
    el = Sphere(1.0, 1.0)
    el2 = el.translate([0, 1, 2])
    assert not np.allclose(el.center, el2.center)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.io.vasp.chg import *

pytestmark = [pytest.mark.io, pytest.mark.vasp]


@pytest.fixture(scope="module", params=["CHG", "CHGCAR"])
def chg_type(request):
    return request.param


@pytest.fixture(scope="module", params=[np.float64, np.float32])
def np_dtype(request):
    return request.param


def test_graphene_chg(sisl_files, chg_type, np_dtype):
    f = sisl_files("vasp", "graphene", chg_type)
    grid = chgSileVASP(f).read_grid(dtype=np_dtype)
    geom = chgSileVASP(f).read_geometry()

    assert grid.grid.sum() * grid.dvolume == pytest.approx(8)
    assert geom == grid.geometry


def test_graphene_chgcar_index_float(sisl_files):
    f = sisl_files("vasp", "graphene", "CHGCAR")
    grid = chgSileVASP(f).read_grid()
    gridh = chgSileVASP(f).read_grid(index=[0.5])

    assert grid.grid.sum() / 2 == pytest.approx(gridh.grid.sum())


@pytest.fixture(scope="module", params=["unpolarized", "polarized", "spin-orbit"])
def spin_conf(request):
    return request.param


def test_nitric_oxide_chg(sisl_files, spin_conf, chg_type, np_dtype):
    f = sisl_files("vasp", "nitric_oxide", spin_conf, chg_type)
    grid = chgSileVASP(f).read_grid(dtype=np_dtype)
    geom = chgSileVASP(f).read_geometry()

    assert grid.grid.sum() * grid.dvolume == pytest.approx(11)
    assert geom == grid.geometry


def test_nitric_oxide_pol(sisl_files, chg_type, np_dtype):
    f = sisl_files("vasp", "nitric_oxide", "polarized", chg_type)
    grid = chgSileVASP(f).read_grid(1, dtype=np_dtype)
    geom = chgSileVASP(f).read_geometry()

    assert grid.grid.sum() * grid.dvolume == pytest.approx(1, rel=1e-3)
    assert geom == grid.geometry
    # check against first raw datapoint on grid
    if chg_type == "CHG":
        assert grid.grid[0, 0, 0] * grid.volume == pytest.approx(0.00045343)
    elif chg_type == "CHGCAR":
        assert grid.grid[0, 0, 0] * grid.volume == pytest.approx(-0.00080296862896)

    # construct up-spin density
    chg = chgSileVASP(f).read_grid(0, dtype=np_dtype)
    grid.grid += chg.grid
    grid.grid /= 2
    up_spin = chgSileVASP(f).read_grid([0.5, 0.5], dtype=np_dtype)
    assert np.allclose(grid.grid, up_spin.grid)


def test_nitric_oxide_soi(sisl_files, chg_type, np_dtype):
    f = sisl_files("vasp", "nitric_oxide", "spin-orbit", chg_type)
    grids = []
    if chg_type == "CHG":
        val = [0.126, 0.00013571, 0.00014291, 0.00015183]
    elif chg_type == "CHGCAR":
        val = [0.13805890712, -0.00058291925321, -0.00059448387648, -0.00060228376205]

    for i in range(4):
        grid = chgSileVASP(f).read_grid(i, dtype=np_dtype)
        grids.append(grid)

        # check against first raw datapoint on grid
        assert grid.grid[0, 0, 0] * grid.volume == pytest.approx(val[i])

    s = grids[0].grid.sum() * grid.dvolume
    assert s == pytest.approx(11, rel=1e-3)

    s = 0
    for i in range(1, 4):
        grid = grids[i]
        s += (grid.grid.sum() * grid.dvolume) ** 2
    assert s == pytest.approx(1, rel=1e-3)

    # construct up-spin density
    grid.grid += grids[0].grid
    grid.grid /= 2
    up_spin = chgSileVASP(f).read_grid([0.5, 0, 0, 0.5], dtype=np_dtype)
    assert np.allclose(grid.grid, up_spin.grid)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

import sisl

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_eigenmode_read(sisl_files):
    fdf = sisl.get_sile(sisl_files("siesta", "H_chain", "h_chain_vibra.fdf"))
    geometry = fdf.read_geometry()
    vectors = sisl.io.siesta.vectorsSileSiesta(
        sisl_files("siesta", "H_chain", "h_chain_vibra.vectors"),
        geometry=geometry,
    )

    blines = fdf.get("BandLines")
    nk = sum(int(l.split()[0]) for l in blines)
    nlines = len(blines)
    iklast = [sum(int(l.split()[0]) for l in blines[: i + 1]) for i in range(nlines)]
    klast = [list(map(float, l.split()[1:4])) for l in blines]

    # yield modes
    nmodes = 0
    nmodes_total = 0
    ks = []
    for state in vectors.yield_eigenmode():
        nmodes += 1
        nmodes_total += len(state)
        ks.append(state.info["k"])

    assert nmodes == nk
    assert nmodes_total == geometry.na * 3 * nk

    bz = vectors.read_brillouinzone()
    assert len(bz) == nk
    assert np.allclose(bz.k, ks)


def test_eigenmode_values(sisl_files):
    fdf = sisl.get_sile(sisl_files("siesta", "H_chain", "h_chain_vibra.fdf"))
    vectors = sisl.io.siesta.vectorsSileSiesta(
        sisl_files("siesta", "H_chain", "h_chain_vibra.vectors"),
        geometry=fdf.read_geometry(),
    )

    mode = vectors.read_eigenmode()
    assert np.allclose(mode.state[0, 0:3], [-0.7071, -0.2400e-4, -0.2400e-4])
    assert np.allclose(mode.state[0, 3:7], [0.7071, 0.2400e-4, 0.2400e-4])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import sys

import numpy as np
import pytest

from sisl._help import (
    array_fill_repeat,
    array_replace,
    dtype_complex_to_float,
    get_dtype,
)

pytestmark = pytest.mark.help


def test_array_fill_repeat1():
    assert array_fill_repeat([1], 20).shape[0] == 20
    assert array_fill_repeat([1, 2], 20).shape[0] == 20
    assert array_fill_repeat(1, 20).shape[0] == 20
    assert array_fill_repeat([[1]], 20, axis=0).shape == (20, 1)
    assert array_fill_repeat([[1, 2]], 20, axis=0).shape == (20, 2)


def test_array_fill_repeat2():
    with pytest.raises(ValueError):
        array_fill_repeat([1, 2, 3], 20)


@pytest.mark.xfail(
    sys.platform.startswith("win"), reason="Datatype cannot be int64 on windows"
)
def test_get_dtype1():
    assert np.int32 == get_dtype(1)
    assert np.int64 == get_dtype(1, int=np.int64)


@pytest.mark.xfail(
    sys.platform.startswith("win"), reason="Datatype cannot be int64 on windows"
)
def test_dtype_complex_to_float():
    for d in (np.int32, np.int64, np.float32, np.float64):
        assert dtype_complex_to_float(d) == d
    assert dtype_complex_to_float(np.complex64) == np.float32
    assert dtype_complex_to_float(np.complex128) == np.float64


def test_array_replace():
    ar = np.arange(6)

    arnew = array_replace(ar, (1, 2))
    assert arnew[1] == 2
    arnew[1] -= 1
    assert np.all(arnew == ar)

    arnew = array_replace(ar, ([1, 3, 5], None), other=4)
    assert np.all(arnew[[1, 3, 5]] == [1, 3, 5])
    assert np.all(np.delete(arnew, [1, 3, 5]) == 4)

    arnew = array_replace(ar, ([1, 3], None), (5, None), other=4)
    assert np.all(arnew[[1, 3, 5]] == [1, 3, 5])
    assert np.all(np.delete(arnew, [1, 3, 5]) == 4)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.shape import *

pytestmark = pytest.mark.shape


def test_create_cuboid():
    cube = Cuboid([1.0] * 3)
    cube = Cuboid([1.0] * 3, [1.0] * 3)
    cube = Cuboid([1.0, 2.0, 3.0], [1.0] * 3)
    cube = Cuboid([1.0, 2.0, 3.0], origin=[1.0] * 3)
    v0 = [1.0, 0.2, 1.0]
    v1 = [1.0, -0.2, 1.0]
    v2 = [1.0, -0.2, -1.0]
    cube = Cuboid([v0, v1, v2])
    str(cube)


def test_create_fail():
    v0 = [1.0, 0.2, 1.0]
    v1 = [1.0, 0.2, 1.0]
    v2 = [1.0, -0.2, -1.0]
    v3 = [1.0, -0.2, -1.0]
    with pytest.raises(ValueError):
        el = Cuboid([v0, v1, v2, v3])
    with pytest.raises(ValueError):
        el = Cuboid(2, center=v1, origin=v2)


def test_tosphere():
    cube = Cube(1.0)
    assert cube.to.Sphere().radius == pytest.approx(0.5 * 3**0.5)
    cube = Cube(3.0)
    assert cube.to.Sphere().radius == pytest.approx(1.5 * 3**0.5)
    cube = Cuboid([1.0, 2.0, 3.0])
    assert cube.to.Sphere().radius == pytest.approx(1.5 * 3**0.5)
    assert cube.to.Sphere().radius == pytest.approx(1.5 * 3**0.5)
    assert isinstance(cube.to.Sphere(), Sphere)


def test_toellipsoid():
    cube = Cube(1.0)
    assert cube.to.Ellipsoid().radius[0] == pytest.approx(0.5 * 3**0.5)
    cube = Cube(3.0)
    assert cube.to.Ellipsoid().radius[0] == pytest.approx(1.5 * 3**0.5)
    cube = Cuboid([1.0, 2.0, 3.0])
    assert cube.to.Ellipsoid().radius[0] == pytest.approx(0.5 * 3**0.5)
    assert cube.to.Ellipsoid().radius[1] == pytest.approx(1 * 3**0.5)
    assert cube.to.Ellipsoid().radius[2] == pytest.approx(1.5 * 3**0.5)


def test_create_cube():
    cube = Cube(1.0)
    cube = Cube(1.0, [1.0] * 3)
    assert cube.volume == pytest.approx(1.0)
    assert cube.scale(2).volume == pytest.approx(2**3)
    assert cube.scale([2] * 3).volume == pytest.approx(2**3)
    assert cube.expand(2).volume == pytest.approx(3**3)
    assert cube.expand([2] * 3).volume == pytest.approx(3**3)


def test_expand_fail():
    cube = Cube(1.0)
    with pytest.raises(ValueError):
        cube.expand([2, 1])


def test_vol1():
    cube = Cuboid([1.0] * 3)
    assert cube.volume == 1.0
    cube = Cuboid([1.0, 2.0, 3.0])
    assert cube.volume == 6.0

    return
    a = (1.0 / 3) ** 0.5
    v0 = [a, a, 0]
    v1 = [-a, a, 0]
    v2 = [0, 0, a]
    cube = Cuboid([v0, v1, v2])
    assert cube.volume == 1.0


def test_origin():
    cube = Cuboid([1.0] * 3)
    assert np.allclose(cube.origin, -0.5)
    cube.origin = 1
    assert np.allclose(cube.origin, 1)


def test_within1():
    cube = Cuboid([1.0] * 3)
    assert not cube.within([-1.0] * 3).any()
    assert not cube.within([[-1.0] * 3, [-1.0, 0.5, 0.2]]).any()
    assert cube.within([[-1.0] * 3, [-1.0, 0.5, 0.2], [0.1, 0.5, 0.2]]).any()


def test_within_index1():
    cube = Cuboid([1.0] * 3)
    assert cube.within_index([-1.0] * 3).size == 0
    assert cube.within_index([[-1.0] * 3, [-1.0, 0.5, 0.2]]).size == 0
    assert (
        cube.within_index([[-1.0] * 3, [-1.0, 0.5, 0.2], [0.1, 0.5, 0.2]]) == [0, 1, 2]
    ).any()


def test_translate():
    el = Cuboid(1.0, 1.0)
    el2 = el.translate([0, 1, 2])
    assert not np.allclose(el.center, el2.center)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

import sisl

pytestmark = [pytest.mark.io, pytest.mark.siesta]

listify = sisl.utils.listify


@pytest.mark.parametrize("sort", [True, False])
@pytest.mark.parametrize(
    "matrix,ext",
    (map(lambda x: ("Hamiltonian", x), ["nc", "TSHS"]) | listify)
    + (map(lambda x: ("DensityMatrix", x), ["nc", "DM"]) | listify)
    + (map(lambda x: ("EnergyDensityMatrix", x), ["nc"]) | listify),
)
@pytest.mark.parametrize("read_dtype", [np.float64, np.complex128])
@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex128])
def test_non_colinear(sisl_tmp, sort, matrix, ext, dtype, read_dtype):
    if ext == "nc":
        pytest.importorskip("netCDF4")

    M = getattr(sisl, matrix)(sisl.geom.graphene(), spin=sisl.Spin("NC"), dtype=dtype)
    if np.issubdtype(dtype, np.complexfloating):
        onsite = [0.1 + 0j, 0.2 + 0j, 0.3 + 0.4j]
        nn = [0.2, 0.3, 0.4 + 0.5j]
    else:
        onsite = [0.1, 0.2, 0.3, 0.4]
        nn = [0.2, 0.3, 0.4, 0.5]
    M.construct(([0.1, 1.44], [onsite, nn]))

    f1 = sisl_tmp(f"M1.{ext}")
    f2 = sisl_tmp(f"M2.{ext}")
    M.write(f1, sort=sort)
    M.finalize()
    with sisl.get_sile(f1) as sile:
        M2 = M.read(sile, dtype=read_dtype)
    M2.write(f2, sort=sort)
    with sisl.get_sile(f2) as sile:
        M3 = M2.read(sile, dtype=read_dtype)

    if sort:
        M.finalize(sort=sort)
    assert M._csr.spsame(M2._csr)
    assert M._csr.spsame(M3._csr)

    # Convert to the same dtype
    M2 = M2.astype(dtype)
    M3 = M3.astype(dtype)
    if M.orthogonal and not M2.orthogonal:
        assert np.allclose(M._csr._D, M2._csr._D[..., :-1])
    else:
        assert np.allclose(M._csr._D, M2._csr._D)
    if M.orthogonal and not M3.orthogonal:
        assert np.allclose(M._csr._D, M3._csr._D[..., :-1])
    else:
        assert np.allclose(M._csr._D, M3._csr._D)


@pytest.mark.parametrize("sort", [True, False])
@pytest.mark.parametrize(
    "matrix,ext",
    (map(lambda x: ("Hamiltonian", x), ["nc", "TSHS"]) | listify)
    + (map(lambda x: ("DensityMatrix", x), ["nc", "DM"]) | listify)
    + (map(lambda x: ("EnergyDensityMatrix", x), ["nc"]) | listify),
)
@pytest.mark.parametrize("read_dtype", [np.float64, np.complex128])
@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex128])
def test_spin_orbit(sisl_tmp, sort, matrix, ext, dtype, read_dtype):
    if ext == "nc":
        pytest.importorskip("netCDF4")

    M = getattr(sisl, matrix)(sisl.geom.graphene(), spin=sisl.Spin("SO"), dtype=dtype)
    if np.issubdtype(dtype, np.complexfloating):
        onsite = [0.1 + 0j, 0.2 + 0j, 0.3 + 0.4j, 0.3 - 0.4j]
        nn = [0.2 + 0.1j, 0.3 + 0.3j, 0.4 + 0.5j, 0.4 - 0.5j]
    else:
        onsite = [0.1, 0.2, 0.3, 0.4, 0, 0, 0.3, -0.4]
        nn = [0.2, 0.3, 0.4, 0.5, 0.1, 0.3, 0.4, -0.5]
    M.construct(([0.1, 1.44], [onsite, nn]))

    f1 = sisl_tmp(f"M1.{ext}")
    f2 = sisl_tmp(f"M2.{ext}")
    M.write(f1, sort=sort)
    M.finalize()
    with sisl.get_sile(f1) as sile:
        M2 = M.read(sile, dtype=read_dtype)
    M2.write(f2, sort=sort)
    with sisl.get_sile(f2) as sile:
        M3 = M2.read(sile, dtype=read_dtype)

    if sort:
        M.finalize(sort=sort)
    assert M._csr.spsame(M2._csr)
    assert M._csr.spsame(M3._csr)

    # Convert to the same dtype
    M2 = M2.astype(dtype)
    M3 = M3.astype(dtype)
    if M.orthogonal and not M2.orthogonal:
        assert np.allclose(M._csr._D, M2._csr._D[..., :-1])
    else:
        assert np.allclose(M._csr._D, M2._csr._D)
    if M.orthogonal and not M3.orthogonal:
        assert np.allclose(M._csr._D, M3._csr._D[..., :-1])
    else:
        assert np.allclose(M._csr._D, M3._csr._D)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import argparse

import pytest

from sisl.utils.cmd import *

pytestmark = pytest.mark.utils


def test_default_namespace1():
    d = {"a": 1}
    dd = default_namespace(**d)
    assert dd.a == d["a"]


def test_collect_input1():
    argv = ["test.xyz", "--stneohus stnaoeu", "-a", "aote"]
    argv_out, in_file = collect_input(argv)

    assert in_file == "test.xyz"
    assert len(argv) == len(argv_out) + 1


def test_collect_arguments1():
    ap, ns, argv = collect_arguments([])
    assert len(argv) == 0


def test_collect_arguments2():
    ap, ns, argv = collect_arguments([], input=True)
    assert len(argv) == 0


def test_collect_arguments3():
    with pytest.raises(ValueError):
        collect_arguments(["this.file.never.exists"], input=True)


def test_decorators1():
    # Create a default argument parser
    @default_ArgumentParser("SPBS", description="MY DEFAULT STUFF")
    def myArgParser(self, p=None, *args, **kwargs):
        return p

    p = myArgParser(None)
    assert "SPBS" in p.format_help()
    assert "MY DEFAULT STUFF" in p.format_help()

    p = argparse.ArgumentParser(description="SECOND DEFAULT")
    p = myArgParser(None, p)
    assert "SPBS" not in p.format_help()
    assert "MY DEFAULT STUFF" in p.format_help()
    assert "SECOND DEFAULT" not in p.format_help()


def test_decorators2():
    p = argparse.ArgumentParser()
    ns = default_namespace(my_default="test")

    class Act1(argparse.Action):
        @run_collect_action
        def __call__(self, parser, ns, value, option_string=None):
            setattr(ns, "act1", value)

    class Act2(argparse.Action):
        @run_collect_action
        def __call__(self, parser, ns, value, option_string=None):
            assert ns.act1 is not None
            setattr(ns, "act2", value)

    class Act3(argparse.Action):
        @collect_action
        def __call__(self, parser, ns, value, option_string=None):
            setattr(ns, "act3", value)

    class Act4(argparse.Action):
        def __call__(self, parser, ns, value, option_string=None):
            with pytest.raises(AttributeError):
                assert ns.act3 is None
            setattr(ns, "act4", value)

    class Act5(argparse.Action):
        @run_actions
        def __call__(self, parser, ns, value, option_string=None):
            pass

    p.add_argument("--a1", action=Act1)
    p.add_argument("--a2", action=Act2)
    p.add_argument("--a3", action=Act3)
    p.add_argument("--a4", action=Act4)
    p.add_argument("--a5", action=Act5, nargs=0)

    # Run arguments
    argv = "--a1 v1 --a2 v2 --a3 v3 --a4 v4".split()
    args = p.parse_args(argv, namespace=ns)

    assert args.my_default == "test"

    assert args.act1 == "v1"
    assert args.act2 == "v2"
    with pytest.raises(AttributeError):
        assert args.act3 is None
    assert args.act4 == "v4"

    args = p.parse_args(argv + ["--a5"], namespace=ns)

    assert args.my_default == "test"

    assert args.act1 == "v1"
    assert args.act2 == "v2"
    assert args.act3 == "v3"
    assert args.act4 == "v4"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import Spin

pytestmark = [pytest.mark.physics, pytest.mark.spin]


def test_spin_init():
    for val in [
        "unpolarized",
        "",
        Spin.UNPOLARIZED,
        "polarized",
        "p",
        Spin.POLARIZED,
        "non-collinear",
        "nc",
        Spin.NONCOLINEAR,
        "spin-orbit",
        "so",
        Spin.SPINORBIT,
        "nambu",
        "bdg",
        Spin.NAMBU,
    ]:
        s = Spin(val)
        str(s)
        s1 = s.copy()
        assert s == s1


def test_spin_comparisons():
    s1 = Spin()
    s2 = Spin("p")
    s3 = Spin("nc")
    s4 = Spin("so")
    s5 = Spin("nambu")

    assert s1.kind == Spin.UNPOLARIZED
    assert s2.kind == Spin.POLARIZED
    assert s3.kind == Spin.NONCOLINEAR
    assert s4.kind == Spin.SPINORBIT
    assert s5.kind == Spin.NAMBU

    assert s1 == s1.copy()
    assert s2 == s2.copy()
    assert s3 == s3.copy()
    assert s4 == s4.copy()
    assert s5 == s5.copy()

    assert s1 < s2
    assert s2 < s3
    assert s3 < s4
    assert s4 < s5

    assert s1 <= s2
    assert s2 <= s3
    assert s3 <= s4

    assert s2 > s1
    assert s3 > s2
    assert s4 > s3
    assert s5 > s4

    assert s2 >= s1
    assert s3 >= s2
    assert s4 >= s3
    assert s5 >= s4

    assert s1.is_unpolarized
    assert not s1.is_polarized
    assert not s1.is_noncolinear
    assert not s1.is_spinorbit
    assert not s1.is_nambu

    assert not s2.is_unpolarized
    assert s2.is_polarized
    assert not s2.is_noncolinear
    assert not s2.is_spinorbit
    assert not s2.is_nambu

    assert not s3.is_unpolarized
    assert not s3.is_polarized
    assert s3.is_noncolinear
    assert not s3.is_spinorbit
    assert not s3.is_nambu

    assert not s4.is_unpolarized
    assert not s4.is_polarized
    assert not s4.is_noncolinear
    assert s4.is_spinorbit
    assert not s4.is_nambu

    assert not s5.is_unpolarized
    assert not s5.is_polarized
    assert not s5.is_noncolinear
    assert not s5.is_spinorbit
    assert s5.is_nambu


def test_spin_unaccepted_arg():
    with pytest.raises(ValueError):
        s = Spin("satoehus")


def test_spin_pauli():
    # just grab the default spin
    S = Spin()

    # Create a fictituous wave-function
    sq2 = 2**0.5
    W = np.array(
        [
            [1 / sq2, 1 / sq2],  # M_x = 1
            [1 / sq2, -1 / sq2],  # M_x = -1
            [0.5 + 0.5j, 0.5 + 0.5j],  # M_x = 1
            [0.5 - 0.5j, -0.5 + 0.5j],  # M_x = -1
            [1 / sq2, 1j / sq2],  # M_y = 1
            [1 / sq2, -1j / sq2],  # M_y = -1
            [0.5 - 0.5j, 0.5 + 0.5j],  # M_y = 1
            [0.5 + 0.5j, 0.5 - 0.5j],  # M_y = -1
            [1, 0],  # M_z = 1
            [0, 1],  # M_z = -1
        ]
    )
    x = np.array([1, -1, 1, -1, 0, 0, 0, 0, 0, 0])
    assert np.allclose(x, (np.conj(W) * S.X.dot(W.T).T).sum(1).real)
    y = np.array([0, 0, 0, 0, 1, -1, 1, -1, 0, 0])
    assert np.allclose(y, (np.conj(W) * np.dot(S.Y, W.T).T).sum(1).real)
    z = np.array([0, 0, 0, 0, 0, 0, 0, 0, 1, -1])
    assert np.allclose(z, (np.conj(W) * np.dot(S.Z, W.T).T).sum(1).real)


def test_spin_pickle():
    import pickle as p

    S = Spin("nc")
    n = p.dumps(S)
    s = p.loads(n)
    assert S == s



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import Geometry, Grid
from sisl.io.xsf import *

pytestmark = [pytest.mark.io, pytest.mark.generic]


def test_xsf_default(sisl_tmp):
    f = sisl_tmp("GRID_default.xsf")
    grid = Grid(0.2)
    grid.grid = np.random.rand(*grid.shape)
    grid.write(f)
    assert grid.geometry is None


@pytest.mark.parametrize(
    "pbc",
    [
        (True, True, True),
        (True, True, False),
        (True, False, False),
        (False, False, False),
    ],
)
def test_xsf_pbc(sisl_tmp, pbc):
    f = sisl_tmp("GRID_default.xsf")
    geom = Geometry(
        np.random.rand(10, 3),
        np.random.randint(1, 70, 10),
        lattice=[10, 10, 10, 45, 60, 90],
    )
    geom.lattice.pbc = pbc
    geom.write(f)
    geom2 = geom.read(f)
    assert all(geom.pbc == geom2.pbc)


def test_xsf_default_size(sisl_tmp):
    f = sisl_tmp("GRID_default_size.xsf")
    grid = Grid(0.2, lattice=2.0)
    grid.grid = np.random.rand(*grid.shape)
    grid.write(f)
    assert grid.geometry is None


def test_xsf_geometry(sisl_tmp):
    f = sisl_tmp("GRID_geometry.xsf")
    geom = Geometry(
        np.random.rand(10, 3),
        np.random.randint(1, 70, 10),
        lattice=[10, 10, 10, 45, 60, 90],
    )
    grid = Grid(0.2, geometry=geom)
    grid.grid = np.random.rand(*grid.shape)
    grid.write(f)
    assert not grid.geometry is None


def test_xsf_imaginary(sisl_tmp):
    f = sisl_tmp("GRID_imag.xsf")
    geom = Geometry(
        np.random.rand(10, 3),
        np.random.randint(1, 70, 10),
        lattice=[10, 10, 10, 45, 60, 90],
    )
    grid = Grid(0.2, geometry=geom, dtype=np.complex128)
    grid.grid = np.random.rand(*grid.shape) + 1j * np.random.rand(*grid.shape)
    grid.write(f)
    assert not grid.geometry is None


def test_axsf_geoms(sisl_tmp):
    f = sisl_tmp("multigeom_nodata.axsf")
    geom = Geometry(
        np.random.rand(10, 3),
        np.random.randint(1, 70, 10),
        lattice=[10, 10, 10, 45, 60, 90],
    )
    geoms = [geom.move((i / 10, i / 10, i / 10)) for i in range(3)]

    with xsfSile(f, "w", steps=3) as s:
        for g in geoms:
            s.write_geometry(g)

    with xsfSile(f) as s:
        rgeoms = s.read_geometry[0:2]()
        assert len(rgeoms) == 2
        assert all(isinstance(rg, Geometry) for rg in rgeoms)
        assert all(g.equal(rg) for g, rg in zip(geoms, rgeoms))

    with xsfSile(f) as s:
        rgeoms = s.read_geometry[1]()
        assert isinstance(rgeoms, Geometry)
        assert geoms[1].equal(rgeoms)

    with xsfSile(f) as s:
        rgeoms = s.read_geometry[:]()
        assert len(rgeoms) == len(geoms)
        assert all(isinstance(rg, Geometry) for rg in rgeoms)
        assert all(g.equal(rg) for g, rg in zip(geoms, rgeoms))

    with xsfSile(f) as s:
        rgeoms, rdata = s.read_geometry[:](ret_data=True)
        assert len(rgeoms) == 3
        assert all(g.equal(rg) for g, rg in zip(geoms, rgeoms))
        for dat in rdata:
            assert dat.size == 0


def test_axsf_data(sisl_tmp):
    f = sisl_tmp("multigeom_data.axsf")
    geom = Geometry(
        np.random.rand(10, 3),
        np.random.randint(1, 70, 10),
        lattice=[10, 10, 10, 45, 60, 90],
    )
    geoms = [geom.move((i / 10, i / 10, i / 10)) for i in range(3)]
    data = np.random.rand(3, 10, 3)

    with xsfSile(f, "w", steps=3) as s:
        for g, dat in zip(geoms, data):
            s.write_geometry(g, data=dat)

    with xsfSile(f) as s:
        rgeoms, rdata = s.read_geometry[:](ret_data=True)
        assert len(rgeoms) == len(geoms)
        assert all(g.equal(rg) for g, rg in zip(geoms, rgeoms))
        assert all(np.allclose(d0, d1) for d0, d1 in zip(rdata, data))

    with xsfSile(f) as s:
        rgeoms, rdata = s.read_geometry[0](ret_data=True)
        assert geoms[0].equal(rgeoms)
        assert np.allclose(rdata, data[0])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

""" pytest test configures """


import numpy as np
import pytest

import sisl

pytestmark = [pytest.mark.io, pytest.mark.gulp]


def test_zz_dynamical_matrix(sisl_files):
    si = sisl.get_sile(sisl_files("gulp", "graphene_zz", "zz.gout"))
    D1 = si.read_dynamical_matrix(order=["got"], cutoff=1.0e-4)
    D2 = si.read_dynamical_matrix(order=["FC"], cutoff=1.0e-4)

    assert D1._csr.spsame(D2._csr)
    D1.finalize()
    D2.finalize()
    assert np.allclose(D1._csr._D, D2._csr._D, atol=1e-5)


def test_zz_sc_geom(sisl_files):
    si = sisl.get_sile(sisl_files("gulp", "graphene_zz", "zz.gout"))
    lattice = si.read_lattice()
    geom = si.read_geometry()
    assert lattice == geom.lattice


def test_graphene_8x8_untiling(sisl_files):
    # Test untiling the graphene example 8 times
    # thanks to Xabier de Cerio for this example
    si = sisl.get_sile(sisl_files("gulp", "ancient", "graphene_8x8.gout"))
    dyn = si.read_dynamical_matrix()

    # Now untile it for different segments
    segs_y = [dyn.untile(8, 1, segment=seg) for seg in range(8)]
    segs_x = [dyn.untile(8, 0, segment=seg) for dyn, seg in zip(segs_y, range(8))]
    for dyn in segs_y:
        dyn.finalize()
    for dyn in segs_x:
        dyn.finalize()

    seg_y = segs_y.pop()
    for seg in segs_y:
        d = seg_y - seg
        assert np.allclose(d._csr._D, 0, atol=1e-6)
        # we can't assert that the sparsity patterns are the same
        # The differences are tiny, 1e-7 but they are there
        # assert seg_y.spsame(seg)

    seg_x = segs_x.pop()
    for seg in segs_x:
        d = seg_x - seg
        assert np.allclose(d._csr._D, 0, atol=1e-6)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import Atom, Geometry, Lattice, SphericalOrbital
from sisl.physics.overlap import Overlap

pytestmark = [pytest.mark.physics, pytest.mark.overlap]


@pytest.fixture
def setup():
    class t:
        def __init__(self):
            bond = 1.42
            sq3h = 3.0**0.5 * 0.5
            self.lattice = Lattice(
                np.array(
                    [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
                )
                * bond,
                nsc=[3, 3, 1],
            )

            n = 60
            rf = np.linspace(0, bond * 1.01, n)
            rf = (rf, rf)
            orb = SphericalOrbital(1, rf, 2.0)
            C = Atom(6, orb.toAtomicOrbital())
            self.g = Geometry(
                np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
                atoms=C,
                lattice=self.lattice,
            )
            self.S = Overlap(self.g)

    return t()


class TestOverlap:
    def test_objects(self, setup):
        assert len(setup.S.xyz) == 2
        assert setup.g.no == len(setup.S)

    def test_dtype(self, setup):
        assert setup.S.dtype == np.float64

    def test_ortho(self, setup):
        assert setup.S.orthogonal

    def test_set1(self, setup):
        S = setup.S.copy()
        S.S[0, 0] = 1.0
        assert S[0, 0] == 1.0
        assert S[1, 0] == 0.0

    def test_fromsp(self, setup):
        S = setup.S.copy()
        csr = S.tocsr()
        S2 = Overlap.fromsp(S.geometry, csr)
        assert np.allclose(csr.data, S2.tocsr().data)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

from sisl import Spin
from sisl.viz.data import BandsData
from sisl.viz.processors.bands import (
    calculate_gap,
    draw_gap,
    draw_gaps,
    filter_bands,
    get_gap_coords,
    sanitize_k,
    style_bands,
)

pytestmark = [pytest.mark.viz, pytest.mark.processors]


@pytest.fixture(
    scope="module", params=["unpolarized", "polarized", "noncolinear", "spinorbit"]
)
def spin(request):
    return Spin(request.param)


@pytest.fixture(scope="module")
def gap():
    return 2.5


@pytest.fixture(scope="module")
def bands_data(spin, gap):
    return BandsData.toy_example(spin=spin, gap=gap)


@pytest.fixture(scope="module", params=["x", "y"])
def E_axis(request):
    return request.param


def test_filter_bands(bands_data):
    spin = bands_data.attrs["spin"]

    # Check that it works without any arguments
    filtered_bands = filter_bands(bands_data)

    # Test filtering by band index.
    filtered_bands = filter_bands(bands_data, bands_range=[0, 5])
    assert np.all(filtered_bands.band == np.arange(0, 6))

    # Test filtering by energy. First check that we actually
    # have bands beyond the energy range that we want to test.
    assert bands_data.E.min() <= -5
    assert bands_data.E.max() >= 5

    filtered_bands = filter_bands(bands_data, Erange=[-5, 5])

    assert filtered_bands.E.min() >= -5
    assert filtered_bands.E.max() <= 5

    if spin.is_polarized:
        filtered_bands = filter_bands(bands_data, Erange=[-5, 5], spin=1)

        assert filtered_bands.E.min() >= -5
        assert filtered_bands.E.max() <= 5

        assert filtered_bands.spin == 1


def test_calculate_gap(bands_data, gap):
    spin = bands_data.attrs["spin"]

    gap_info = calculate_gap(bands_data)

    # Check that the gap value is correct
    assert gap_info["gap"] == gap

    # Check also that the position of the gap is in the information
    assert isinstance(gap_info["k"], tuple) and len(gap_info["k"]) == 2

    VB = len(bands_data.band) // 2 - 1
    assert isinstance(gap_info["bands"], tuple) and len(gap_info["bands"]) == 2
    assert gap_info["bands"][0] < gap_info["bands"][1]

    assert isinstance(gap_info["spin"], tuple) and len(gap_info["spin"]) == 2
    if not spin.is_polarized:
        assert gap_info["spin"] == (0, 0)

    assert isinstance(gap_info["Es"], tuple) and len(gap_info["Es"]) == 2
    assert np.allclose(gap_info["Es"], (-gap / 2, gap / 2))


def test_sanitize_k(bands_data):
    assert sanitize_k(bands_data, "Gamma") == 0
    assert sanitize_k(bands_data, "X") == 1


def test_get_gap_coords(bands_data):
    spin = bands_data.attrs["spin"]

    vb = len(bands_data.band) // 2 - 1

    # We can get the gamma gap by specifying both origin and destination or
    # just origin. Check also Gamma to X.
    for to_k in ["Gamma", None, "X"]:
        k, E = get_gap_coords(
            bands_data, (vb, vb + 1), from_k="Gamma", to_k=to_k, spin=1
        )

        kval = 1 if to_k == "X" else 0

        # Check that the K is correct
        assert k[0] == 0
        assert k[1] == kval

        # Check that E is correct.
        if spin.is_polarized:
            bands_E = bands_data.E.sel(spin=1)
        else:
            bands_E = bands_data.E

        assert E[0] == bands_E.sel(band=vb, k=0)
        assert E[1] == bands_E.sel(band=vb + 1, k=kval)


def test_draw_gap(E_axis):
    ks = (0, 0.5)
    Es = (0, 1)

    if E_axis == "x":
        x, y = Es, ks
    else:
        x, y = ks, Es

    gap_action = draw_gap(ks, Es, color="red", name="test", E_axis=E_axis)

    assert isinstance(gap_action, dict)
    assert gap_action["method"] == "draw_line"

    action_kwargs = gap_action["kwargs"]

    assert action_kwargs["name"] == "test"
    assert action_kwargs["line"]["color"] == "red"
    assert action_kwargs["marker"]["color"] == "red"
    assert action_kwargs["x"] == x
    assert action_kwargs["y"] == y


@pytest.mark.parametrize("display_gap", [True, False])
def test_draw_gaps(bands_data, E_axis, display_gap):
    spin = bands_data.attrs["spin"]

    gap_info = calculate_gap(bands_data)

    # Run the function only to draw the minimum gap.
    gap_actions = draw_gaps(
        bands_data,
        gap=display_gap,
        gap_info=gap_info,
        gap_tol=0.3,
        gap_color="red",
        gap_marker={},
        direct_gaps_only=False,
        custom_gaps=[],
        E_axis=E_axis,
    )

    assert isinstance(gap_actions, list)
    assert len(gap_actions) == (1 if display_gap else 0)

    if display_gap:
        assert isinstance(gap_actions[0], dict)
        assert gap_actions[0]["method"] == "draw_line"

        action_kwargs = gap_actions[0]["kwargs"]
        assert action_kwargs["line"]["color"] == "red"
        assert action_kwargs["marker"]["color"] == "red"

    # Now run the function with a custom gap.
    gap_actions = draw_gaps(
        bands_data,
        gap=display_gap,
        gap_info=gap_info,
        gap_tol=0.3,
        gap_color="red",
        gap_marker={},
        direct_gaps_only=False,
        custom_gaps=[{"from": "Gamma", "to": "X", "color": "blue"}],
        E_axis=E_axis,
    )

    assert isinstance(gap_actions, list)
    assert len(gap_actions) == (2 if display_gap else 1) + (
        1 if spin.is_polarized else 0
    )

    # Check the minimum gap
    if display_gap:
        assert isinstance(gap_actions[0], dict)
        assert gap_actions[0]["method"] == "draw_line"

        action_kwargs = gap_actions[0]["kwargs"]
        assert action_kwargs["line"]["color"] == "red"
        assert action_kwargs["marker"]["color"] == "red"

    # Check the custom gap
    assert isinstance(gap_actions[-1], dict)
    assert gap_actions[-1]["method"] == "draw_line"

    action_kwargs = gap_actions[-1]["kwargs"]
    assert action_kwargs["line"]["color"] == "blue"
    assert action_kwargs["marker"]["color"] == "blue"
    assert action_kwargs["x" if E_axis == "y" else "y"] == (0, 1)


def test_style_bands(bands_data):
    spin = bands_data.attrs["spin"]

    # Check basic styles
    styled_bands = style_bands(
        bands_data,
        {"color": "red", "width": 3},
        spindown_style={"opacity": 0.5, "color": "blue"},
    )

    assert isinstance(styled_bands, xr.Dataset)

    for k in ("color", "width", "opacity"):
        assert k in styled_bands.data_vars

    if not spin.is_polarized:
        assert styled_bands.color == "red"
        assert styled_bands.width == 3
        assert styled_bands.opacity == 1
    else:
        assert np.all(styled_bands.color == ["red", "blue"])
        assert np.all(styled_bands.width == [3, 3])
        assert np.all(styled_bands.opacity == [1, 0.5])

    # Check function as style
    def color(data):
        return xr.DataArray(
            np.where(data.band < 5, "red", "blue"), coords=[("band", data.band.values)]
        )

    styled_bands = style_bands(
        bands_data,
        {"color": color, "width": 3},
        spindown_style={"opacity": 0.5, "color": "blue"},
    )

    assert isinstance(styled_bands, xr.Dataset)

    for k in ("color", "width", "opacity"):
        assert k in styled_bands.data_vars

    assert "band" in styled_bands.color.coords
    if spin.is_polarized:
        bands_color = styled_bands.color.sel(spin=0)
        assert np.all(styled_bands.color.sel(spin=1) == "blue")
    else:
        bands_color = styled_bands.color
    assert np.all((styled_bands.band < 5) == (bands_color == "red"))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest

approx = pytest.approx

from sisl.unit.siesta import unit_convert, unit_default, unit_group

pytestmark = [pytest.mark.unit, pytest.mark.siesta]


def test_group():
    assert unit_group("kg") == "mass"
    assert unit_group("eV") == "energy"
    assert unit_group("N") == "force"


def test_unit_convert():
    assert approx(unit_convert("kg", "g")) == 1.0e3
    assert approx(unit_convert("eV", "J")) == 1.602176634e-19
    assert approx(unit_convert("J", "eV")) == 1 / 1.602176634e-19
    assert approx(unit_convert("J", "eV", {"^": 2})) == (1 / 1.602176634e-19) ** 2
    assert approx(unit_convert("J", "eV", {"/": 2})) == (1 / 1.602176634e-19) / 2
    assert approx(unit_convert("J", "eV", {"*": 2})) == (1 / 1.602176634e-19) * 2


def test_default():
    assert unit_default("mass") == "amu"
    assert unit_default("energy") == "eV"
    assert unit_default("force") == "eV/Ang"


def test_group_f1():
    with pytest.raises(ValueError):
        unit_group("not-existing")


def test_default_f1():
    with pytest.raises(ValueError):
        unit_default("not-existing")


def test_unit_convert_f1():
    with pytest.raises(ValueError):
        unit_convert("eV", "megaerg")


def test_unit_convert_f2():
    with pytest.raises(ValueError):
        unit_convert("eV", "kg")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import itertools
from functools import partial

import numpy as np
import pytest

from sisl import Atom, Lattice, SislError
from sisl._math_small import cross3, dot3
from sisl.geom import *

pytestmark = [pytest.mark.geom]


class CellDirect(Lattice):
    @property
    def volume(self):
        return dot3(self.cell[0, :], cross3(self.cell[1, :], self.cell[2, :]))


def is_right_handed(geometry):
    sc = CellDirect(geometry.lattice.cell)
    return sc.volume > 0.0


def has_vacuum(geometry, vacuum):
    fxyz = geometry.fxyz
    # print(fxyz.max(0) - fxyz.min(0), geometry.lattice.length)
    vac = (1 - (fxyz.max(0) - fxyz.min(0))) * geometry.lattice.length
    vac = vac[~geometry.lattice.pbc]
    print(vac, vacuum)
    return np.allclose(vac, vacuum)


def test_basic():
    a = sc(2.52, Atom["Fe"])
    assert is_right_handed(a)
    a = bcc(2.52, Atom["Fe"])
    assert is_right_handed(a)
    a = bcc(2.52, Atom["Fe"], orthogonal=True)
    a = fcc(2.52, Atom["Au"])
    assert is_right_handed(a)
    a = fcc(2.52, Atom["Au"], orthogonal=True)
    a = hcp(2.52, Atom["Au"])
    assert is_right_handed(a)
    a = hcp(2.52, Atom["Au"], orthogonal=True)
    a = rocksalt(5.64, ["Na", "Cl"])
    assert is_right_handed(a)
    a = rocksalt(5.64, [Atom("Na", R=3), Atom("Cl", R=4)], orthogonal=True)


@pytest.mark.parametrize(
    "func, orthogonal, vacuum",
    itertools.product([graphene, goldene], [True, False], [10, 20]),
)
def test_flat(func, orthogonal, vacuum):
    a = func(orthogonal=orthogonal, vacuum=vacuum)
    assert is_right_handed(a)
    assert has_vacuum(a, vacuum)
    a = func(atoms="C", orthogonal=orthogonal, vacuum=vacuum)
    assert is_right_handed(a)
    assert has_vacuum(a, vacuum)


@pytest.mark.parametrize(
    "vacuum",
    [10, (20, 30, 10)],
)
def test_flat_flakes(vacuum):
    g = graphene_flake(shells=0, bond=1.42, vacuum=vacuum)
    assert g.na == 6
    # All atoms are close to the center
    assert len(g.close(g.center(), 1.44)) == g.na
    # All atoms have two neighbors
    assert len(g.axyz(AtomNeighbors(min=2, max=2, R=1.44))) == g.na
    assert is_right_handed(g)
    assert has_vacuum(g, vacuum)

    g = graphene_flake(shells=1, bond=1.42, vacuum=vacuum)
    assert g.na == 24
    assert len(g.close(g.center(), 4)) == g.na
    assert len(g.axyz(AtomNeighbors(min=2, max=2, R=1.44))) == 12
    assert len(g.axyz(AtomNeighbors(min=3, max=3, R=1.44))) == 12
    assert is_right_handed(g)
    assert has_vacuum(g, vacuum)

    bn = honeycomb_flake(shells=1, atoms=["B", "N"], bond=1.42, vacuum=vacuum)
    assert bn.na == 24
    assert np.allclose(bn.xyz, g.xyz)
    # Check that atoms are alternated.
    assert len(bn.axyz(AtomZ(5) & AtomNeighbors(min=1, R=1.44, neighbor=AtomZ(5)))) == 0
    assert is_right_handed(bn)
    assert has_vacuum(bn, vacuum)


@pytest.mark.parametrize(
    "vacuum",
    [10, (20, 30, 10)],
)
def test_triangulene(vacuum):
    g = triangulene(3, vacuum=vacuum)
    assert g.na == 22
    assert is_right_handed(g)
    assert has_vacuum(g, vacuum)

    g = triangulene(3, atoms=["B", "N"], vacuum=vacuum)
    assert g.atoms.nspecies == 2
    assert is_right_handed(g)
    assert has_vacuum(g, vacuum)


@pytest.mark.parametrize(
    "vacuum",
    [10, (30, 10)],
)
def test_nanotube(vacuum):
    a = nanotube(1.42, vacuum=vacuum)
    assert is_right_handed(a)
    assert has_vacuum(a, vacuum)
    a = nanotube(1.42, chirality=(3, 5), vacuum=vacuum)
    assert is_right_handed(a)
    assert has_vacuum(a, vacuum)
    a = nanotube(1.42, chirality=(6, -3), vacuum=vacuum)
    assert is_right_handed(a)
    assert has_vacuum(a, vacuum)


def test_diamond():
    a = diamond()
    assert is_right_handed(a)


@pytest.mark.parametrize(
    "vacuum",
    [10, 30],
)
def test_bilayer(vacuum):
    a = bilayer(1.42, vacuum=vacuum)
    assert is_right_handed(a)
    assert has_vacuum(a, vacuum)


def test_bilayer_arguments():
    # Just test arguments
    bilayer(1.42, stacking="AA")
    bilayer(1.42, stacking="BA")
    bilayer(1.42, stacking="AB")
    for m in range(7):
        bilayer(1.42, twist=(m, m + 1))
    bilayer(1.42, twist=(6, 7), layer="bottom")
    bilayer(1.42, twist=(6, 7), layer="TOP")
    bilayer(1.42, bottom_atoms=(Atom["B"], Atom["N"]), twist=(6, 7))
    bilayer(1.42, top_atoms=(Atom(5), Atom(7)), twist=(6, 7))
    _, _ = bilayer(1.42, twist=(6, 7), ret_angle=True)

    with pytest.raises(ValueError):
        bilayer(1.42, twist=(6, 7), layer="undefined")

    with pytest.raises(ValueError):
        bilayer(1.42, twist=(6, 7), stacking="undefined")

    with pytest.raises(ValueError):
        bilayer(1.42, twist=("str", 7), stacking="undefined")


@pytest.mark.parametrize(
    "vacuum",
    [10, (30, 10)],
)
def test_nanoribbon(vacuum):
    for w in range(0, 5):
        a = nanoribbon(w, 1.42, (Atom(5), Atom(7)), kind="zigzag", vacuum=vacuum)
        assert is_right_handed(a)
        assert has_vacuum(a, vacuum)


def test_nanoribbon_arguments():
    for w in range(0, 5):
        nanoribbon(w, 1.42, Atom(6), kind="armchair")
        nanoribbon(w, 1.42, Atom(6), kind="zigzag")
        nanoribbon(w, 1.42, Atom(6), kind="chiral")
        nanoribbon(w, 1.42, Atom(6), kind="chiral", chirality=(2, 2))
        nanoribbon(w, 1.42, (Atom(5), Atom(7)), kind="armchair")

    with pytest.raises(ValueError):
        nanoribbon(6, 1.42, (Atom(5), Atom(7)), kind="undefined")

    with pytest.raises(ValueError):
        nanoribbon("str", 1.42, (Atom(5), Atom(7)), kind="undefined")


def test_graphene_nanoribbon():
    graphene_nanoribbon(6, kind="armchair")


@pytest.mark.parametrize(
    "vacuum",
    [10, (30, 10)],
)
def test_agnr(vacuum):
    a = agnr(5, vacuum=vacuum)
    assert is_right_handed(a)
    assert has_vacuum(a, vacuum)


@pytest.mark.parametrize(
    "vacuum",
    [10, (30, 10)],
)
def test_zgnr(vacuum):
    a = zgnr(5, vacuum=vacuum)
    assert is_right_handed(a)
    assert has_vacuum(a, vacuum)


@pytest.mark.parametrize(
    "vacuum",
    [10, (30, 10)],
)
def test_cgnr(vacuum):
    a = cgnr(6, (3, 1), atoms=["B", "N"], vacuum=vacuum)
    assert is_right_handed(a)
    assert has_vacuum(a, vacuum)


@pytest.mark.parametrize(
    "W, invert_first",
    itertools.product(range(3, 20), [True, False]),
)
def test_heteroribbon_one_unit(W, invert_first):
    # Check that all ribbon widths are properly cut into one unit
    geometry = heteroribbon(
        [(W, 1)], bond=1.42, atoms=Atom(6, 1.43), invert_first=invert_first
    )

    assert geometry.na == W


def test_heteroribbon():
    """Runs the heteroribbon builder for all possible combinations of
    widths and asserts that they are always properly aligned.
    """
    # Build combinations
    combinations = itertools.product([7, 8, 9, 10, 11], [7, 8, 9, 10, 11])
    L = itertools.repeat(2)

    for Ws in combinations:
        geom = heteroribbon(
            zip(Ws, L), bond=1.42, atoms=Atom(6, 1.43), align="auto", shift_quantum=True
        )

        # Assert no dangling bonds.
        assert len(geom.asc2uc({"neighbors": 1})) == 0


def test_graphene_heteroribbon():
    a = graphene_heteroribbon([(7, 2), (9, 2)])


def test_graphene_heteroribbon_errors():
    # 7-open with 9 can only be perfectly aligned.
    graphene_heteroribbon([(7, 1), (9, 1)], align="center", on_lone_atom="raise")
    with pytest.raises(SislError):
        graphene_heteroribbon(
            [(7, 1), (9, 1, -1)], align="center", on_lone_atom="raise"
        )
    # From the bottom
    graphene_heteroribbon([(7, 1), (9, 1, -1)], align="bottom", on_lone_atom="raise")
    with pytest.raises(SislError):
        graphene_heteroribbon([(7, 1), (9, 1, 0)], align="bottom", on_lone_atom="raise")
    # And from the top
    graphene_heteroribbon([(7, 1), (9, 1, 1)], align="top", on_lone_atom="raise")
    with pytest.raises(SislError):
        graphene_heteroribbon([(7, 1), (9, 1, -1)], align="top", on_lone_atom="raise")

    grap_heteroribbon = partial(graphene_heteroribbon, align="auto", shift_quantum=True)

    # Odd section with open end
    with pytest.raises(SislError):
        grap_heteroribbon([(7, 3), (5, 2)])

    # Shift limits are imposed correctly
    # In this case -2 < shift < 1
    grap_heteroribbon([(7, 3), (11, 2, 0)])
    grap_heteroribbon([(7, 3), (11, 2, -1)])
    with pytest.raises(SislError):
        grap_heteroribbon([(7, 3), (11, 2, 1)])
    with pytest.raises(SislError):
        grap_heteroribbon([(7, 3), (11, 2, -2)])

    # Periodic boundary conditions work properly
    # grap_heteroribbon([[10, 2], [8, 1, 0]], pbc=False)
    # with pytest.raises(ValueError):
    #     grap_heteroribbon([[10, 2], [8, 1, 0]], pbc=True)

    # Even ribbons should only be shifted towards the center
    grap_heteroribbon([(10, 2), (8, 2, -1)])
    with pytest.raises(SislError):
        grap_heteroribbon([(10, 2), (8, 2, 1)])
    grap_heteroribbon(
        [(10, 1), (8, 2, 1)],
    )  # pbc=False)
    with pytest.raises(SislError):
        grap_heteroribbon(
            [(10, 1), (8, 2, -1)],
        )  # pbc=False)


def test_fcc_slab():
    for o in [True, False]:
        fcc_slab(alat=4.08, atoms="Au", miller=(1, 0, 0), orthogonal=o)
        fcc_slab(4.08, "Au", 100, orthogonal=o)
        fcc_slab(4.08, "Au", 110, orthogonal=o)
        fcc_slab(4.08, "Au", 111, orthogonal=o)
        fcc_slab(4.08, 79, "100", layers=5, vacuum=None, orthogonal=o)
        fcc_slab(4.08, 79, "110", layers=5, orthogonal=o)
        fcc_slab(4.08, 79, "111", layers=5, start=1, orthogonal=o)
        fcc_slab(4.08, 79, "111", layers=5, start="C", orthogonal=o)
        fcc_slab(4.08, 79, "111", layers=5, end=2, orthogonal=o)
        a = fcc_slab(4.08, 79, "111", layers=5, end="B", orthogonal=o)
        assert is_right_handed(a)

    with pytest.raises(ValueError):
        fcc_slab(4.08, "Au", 100, start=0, end=0)
    with pytest.raises(ValueError):
        fcc_slab(4.08, "Au", 1000)
    with pytest.raises(NotImplementedError):
        fcc_slab(4.08, "Au", 200)
    assert not np.allclose(
        fcc_slab(5.64, "Au", 100, end=1, layers="BABAB").xyz,
        fcc_slab(5.64, "Au", 100, end=1, layers=" BABAB ").xyz,
    )
    assert np.allclose(
        fcc_slab(5.64, "Au", 100, layers=" AB AB BA ", vacuum=2).xyz,
        fcc_slab(
            5.64, "Au", 100, layers=(None, 2, 2, 2, None), vacuum=2, start=(0, 0, 1)
        ).xyz,
    )
    assert np.allclose(
        fcc_slab(5.64, "Au", 100, layers=" AB AB BA ", vacuum=(2, 1)).xyz,
        fcc_slab(
            5.64,
            "Au",
            100,
            layers=(None, 2, " ", 2, None, 2, None),
            vacuum=(2, 1),
            end=(1, 1, 0),
        ).xyz,
    )

    # example in documentation
    assert np.allclose(
        fcc_slab(
            4.0, "Au", 100, layers=(" ", 3, 5, 3), start=(0, 1, 0), vacuum=(10, 1, 2)
        ).xyz,
        fcc_slab(4.0, "Au", 100, layers=" ABA BABAB ABA", vacuum=(10, 1, 2)).xyz,
    )

    assert np.allclose(
        fcc_slab(
            4.0, "Au", 100, layers=(" ", 3, 5, 3), start=(1, 0), vacuum=(10, 1, 2)
        ).xyz,
        fcc_slab(4.0, "Au", 100, layers=" BAB ABABA ABA", vacuum=(10, 1, 2)).xyz,
    )


def test_bcc_slab():
    for o in [True, False]:
        bcc_slab(alat=4.08, atoms="Au", miller=(1, 0, 0), orthogonal=o)
        bcc_slab(4.08, "Au", 100, orthogonal=o)
        bcc_slab(4.08, "Au", 110, orthogonal=o)
        bcc_slab(4.08, "Au", 111, orthogonal=o)
        bcc_slab(4.08, 79, "100", layers=5, vacuum=None, orthogonal=o)
        bcc_slab(4.08, 79, "110", layers=5, orthogonal=o)
        assert bcc_slab(4.08, 79, "111", layers=5, start=1, orthogonal=o).equal(
            bcc_slab(4.08, 79, "111", layers="BCABC", orthogonal=o)
        )
        bcc_slab(4.08, 79, "111", layers="BCABC", start="B", orthogonal=o)
        bcc_slab(4.08, 79, "111", layers="BCABC", start=1, orthogonal=o)
        bcc_slab(4.08, 79, "111", layers=5, start="C", orthogonal=o)
        bcc_slab(4.08, 79, "111", layers=5, end=2, orthogonal=o)
        a = bcc_slab(4.08, 79, "111", layers=5, end="B", orthogonal=o)
        assert is_right_handed(a)

    with pytest.raises(ValueError):
        bcc_slab(4.08, "Au", 100, start=0, end=0)
    with pytest.raises(ValueError):
        bcc_slab(4.08, "Au", 1000)
    with pytest.raises(NotImplementedError):
        bcc_slab(4.08, "Au", 200)
    assert not np.allclose(
        bcc_slab(5.64, "Au", 100, end=1, layers="BABAB").xyz,
        bcc_slab(5.64, "Au", 100, end=1, layers=" BABAB ").xyz,
    )


def test_rocksalt_slab():
    rocksalt_slab(5.64, [Atom(11, R=3), Atom(17, R=4)], 100)
    assert rocksalt_slab(5.64, ["Na", "Cl"], 100, layers=5).equal(
        rocksalt_slab(5.64, ["Na", "Cl"], 100, layers="ABABA")
    )
    rocksalt_slab(5.64, ["Na", "Cl"], 110, vacuum=None)
    rocksalt_slab(5.64, ["Na", "Cl"], 111, orthogonal=False)
    rocksalt_slab(5.64, ["Na", "Cl"], 111, orthogonal=True)
    a = rocksalt_slab(5.64, "Na", 100)
    assert is_right_handed(a)

    with pytest.raises(ValueError):
        rocksalt_slab(5.64, ["Na", "Cl"], 100, start=0, end=0)
    with pytest.raises(ValueError):
        rocksalt_slab(5.64, ["Na", "Cl"], 1000)
    with pytest.raises(NotImplementedError):
        rocksalt_slab(5.64, ["Na", "Cl"], 200)
    with pytest.raises(ValueError):
        rocksalt_slab(5.64, ["Na", "Cl"], 100, start=0, layers="BABAB")
    rocksalt_slab(5.64, ["Na", "Cl"], 100, start=1, layers="BABAB")
    with pytest.raises(ValueError):
        rocksalt_slab(5.64, ["Na", "Cl"], 100, end=0, layers="BABAB")
    rocksalt_slab(5.64, ["Na", "Cl"], 100, end=1, layers="BABAB")
    assert not np.allclose(
        rocksalt_slab(5.64, ["Na", "Cl"], 100, end=1, layers="BABAB").xyz,
        rocksalt_slab(5.64, ["Na", "Cl"], 100, end=1, layers=" BABAB ").xyz,
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest

import sisl

pytestmark = pytest.mark.version


def test_version():
    sisl.__bibtex__
    sisl.__version__


def test_import_simple():
    # The imports should only be visible in the io module
    sisl.BaseSile
    sisl.Sile
    sisl.SileCDF
    sisl.SileBin
    sisl.io.xyzSile


def test_submodule_attr_access():
    for mod in (
        "geom",
        "io",
        "physics",
        "linalg",
        "shape",
        "mixing",
        "utils",
        "unit",
        "C",
        "constant",
    ):
        getattr(sisl, mod)


def test_submodule_attr_access_viz():
    pytest.importorskip("plotly")
    sisl.viz


def test_import_in_io():
    # The imports should only be visible in the io module
    with pytest.raises(AttributeError):
        sisl.xyzSile


def test_import_in_io_from():
    # The imports should only be visible in the io module
    with pytest.raises(ImportError):
        from sisl import xyzSile  # noqa: F401


@pytest.mark.parametrize("obj", [dict(), 2])
def test_dispatch_methods_not_allowed(obj):
    with pytest.raises(sisl.SislError):
        sisl.tile(obj, 0, 2)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

""" pytest test configures """


import numpy as np
import pytest

import sisl

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_dm_si_pdos_kgrid(sisl_files):
    fdf = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.fdf"))
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.DM"))

    DM1 = si.read_density_matrix(geometry=fdf.read_geometry())
    DM2 = fdf.read_density_matrix(order=["DM"])

    assert DM1._csr.spsame(DM2._csr)
    assert np.allclose(DM1._csr._D[:, :-1], DM2._csr._D[:, :-1])


def test_dm_si_pdos_kgrid_rw(sisl_files, sisl_tmp):
    fdf = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.fdf"))
    geom = fdf.read_geometry()

    f1 = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.DM"))
    f2 = sisl.get_sile(sisl_tmp("test.DM"))

    DM1 = f1.read_density_matrix(geometry=geom)
    f2.write_density_matrix(DM1, sort=False)
    DM2 = f2.read_density_matrix(geometry=geom)

    assert DM1._csr.spsame(DM2._csr)
    assert np.allclose(DM1._csr._D[:, :-1], DM2._csr._D[:, :-1])

    f2.write_density_matrix(DM1)
    DM2 = f2.read_density_matrix(sort=False)
    assert DM1._csr.spsame(DM2._csr)
    assert not np.allclose(DM1._csr._D[:, :-1], DM2._csr._D[:, :-1])
    DM2.finalize()
    assert np.allclose(DM1._csr._D[:, :-1], DM2._csr._D[:, :-1])


def test_dm_si_pdos_kgrid_mulliken(sisl_files):
    fdf = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.fdf"))
    DM = fdf.read_density_matrix(order=["DM"])

    Mo = DM.mulliken("orbital")
    Ma = DM.mulliken("atom")

    o2a = DM.geometry.o2a(np.arange(DM.no))

    ma = np.zeros_like(Ma)
    np.add.at(ma, o2a, Mo)
    assert np.allclose(ma, Ma)


def test_dm_soc_pt2_xx_mulliken(sisl_files):
    fdf = sisl.get_sile(sisl_files("siesta", "Pt2_soc", "Pt2.fdf"))
    # Force reading a geometry with correct atomic and orbital configuration
    DM = fdf.read_density_matrix(order=["DM"])

    Mo = DM.mulliken("orbital")
    Ma = DM.mulliken("atom")

    o2a = DM.geometry.o2a(np.arange(DM.no))

    ma = np.zeros_like(Ma)
    np.add.at(ma.T, o2a, Mo.T)
    assert np.allclose(ma, Ma)


def test_dm_soc_pt2_xx_rw(sisl_files, sisl_tmp):
    f1 = sisl.get_sile(sisl_files("siesta", "Pt2_soc", "Pt2_xx.DM"))
    f2 = sisl.get_sile(sisl_tmp("test.DM"))

    DM1 = f1.read_density_matrix()
    f2.write_density_matrix(DM1)
    DM2 = f2.read_density_matrix()

    assert DM1._csr.spsame(DM2._csr)
    DM1.finalize()
    assert DM1._csr.spsame(DM2._csr)
    assert np.allclose(DM1._csr._D[:, :-1], DM2._csr._D[:, :-1])


@pytest.mark.xfail(
    reason="Currently reading a geometry from TSHS does not retain l, m, zeta quantum numbers"
)
def test_dm_soc_pt2_xx_orbital_momentum(sisl_files):
    fdf = sisl.get_sile(sisl_files("siesta", "Pt2_soc", "Pt2.fdf"))
    # Force reading a geometry with correct atomic and orbital configuration
    DM = fdf.read_density_matrix(order=["DM"])

    o2a = DM.geometry.o2a(np.arange(DM.no))

    # Calculate angular momentum
    Lo = DM.orbital_momentum("orbital")
    La = DM.orbital_momentum("atom")

    la = np.zeros_like(La)
    np.add.at(la, o2a, Lo.T)
    assert np.allclose(la, La)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.io.siesta.fc import *
from sisl.unit.siesta import unit_convert

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_read_fc(sisl_tmp):
    f = sisl_tmp("test.FC")

    fc = np.random.rand(20, 6, 2, 3)
    sign = 1
    with open(f, "w") as fh:
        fh.write("sotaeuha 2 0.5\n")
        for n in fc:
            for dx in n:
                for a in dx * sign:
                    fh.write("{} {} {}\n".format(*a))
                sign *= -1
    fc.shape = (20, 3, 2, 2, 3)

    # Since the fc file re-creates the sign and divides by the length we have to do this
    fc2 = fcSileSiesta(f).read_force() / 0.5
    assert fc.shape == fc2.shape
    assert np.allclose(fc, fc2)


@pytest.mark.filterwarnings("ignore", message="*assumes displacement=")
def test_read_fc_old(sisl_tmp):
    f = sisl_tmp("test2.FC")

    fc = np.random.rand(20, 6, 2, 3)
    with open(f, "w") as fh:
        fh.write("sotaeuha\n")
        for n in fc:
            for dx in n:
                for a in dx:
                    fh.write("{} {} {}\n".format(*a))
    fc.shape = (20, 3, 2, 2, 3)

    fc2 = fcSileSiesta(f).read_force() / (0.04 * unit_convert("Bohr", "Ang"))
    assert fc.shape != fc2.shape
    fc2 *= np.tile([1, -1], 3).reshape(1, 3, 2, 1, 1)
    fc2.shape = (-1, 3, 2, 2, 3)
    assert fc.shape == fc2.shape
    assert np.allclose(fc, fc2)

    fc2 = fcSileSiesta(f).read_hessian()
    assert fc.shape != fc2.shape
    fc2.shape = (-1, 3, 2, 2, 3)
    assert fc.shape == fc2.shape
    assert np.allclose(fc, fc2)

    # Specify number of atoms
    fc2 = fcSileSiesta(f).read_hessian(2)
    assert fc.shape == fc2.shape
    assert np.allclose(fc, fc2)

    # Specify number of atoms and correction to check they are equivalent
    fc2 = fcSileSiesta(f).read_force(-1.0, na=2)
    assert fc.shape == fc2.shape
    fc2 *= np.tile([-1, 1], 3).reshape(1, 3, 2, 1, 1)
    assert np.allclose(fc, fc2)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import geom
from sisl.io.siesta.kp import *
from sisl.physics import MonkhorstPack

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_kp_read_write(sisl_tmp):
    f = sisl_tmp("tmp.KP")
    g = geom.graphene()
    bz = MonkhorstPack(g, [10, 10, 10])
    kpSileSiesta(f, "w").write_brillouinzone(bz)

    kpoints, weights = kpSileSiesta(f).read_data(g)
    assert np.allclose(kpoints, bz.k)
    assert np.allclose(weights, bz.weight)

    kpoints, weights = kpSileSiesta(f).read_data()
    assert np.allclose(kpoints, bz.tocartesian(bz.k))
    assert np.allclose(weights, bz.weight)

    bz2 = kpSileSiesta(f).read_brillouinzone(g.lattice)
    assert np.allclose(bz2.k, bz.k)
    assert np.allclose(bz2.weight, bz.weight)


def test_rkp_read_write(sisl_tmp):
    f = sisl_tmp("tmp.RKP")
    g = geom.graphene()
    bz = MonkhorstPack(g, [10, 10, 10])
    rkpSileSiesta(f, "w").write_brillouinzone(bz)

    kpoints, weights = rkpSileSiesta(f).read_data()
    assert np.allclose(kpoints, bz.k)
    assert np.allclose(weights, bz.weight)

    bz2 = rkpSileSiesta(f).read_brillouinzone(g.lattice)
    assert np.allclose(bz2.k, bz.k)
    assert np.allclose(bz2.weight, bz.weight)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

import sisl
from sisl.io.siesta.eig import *
from sisl.io.siesta.fdf import *

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def _convert45(unit):
    """Convert from legacy units to CODATA2018"""
    from sisl.unit.siesta import units, units_legacy

    return units(unit) / units_legacy(unit)


def test_si_pdos_kgrid_eig(sisl_files):
    f = sisl_files("siesta", "Si_pdos_k", "Si_pdos.EIG")
    eig = eigSileSiesta(f).read_data()

    # nspin, nk, nb
    assert np.all(eig.shape == (1, 63, 18))


def test_si_pdos_kgrid_eig_ArgumentParser(sisl_files, sisl_tmp):
    pytest.importorskip("matplotlib", reason="matplotlib not available")
    png = sisl_tmp("si_pdos_kgrid.EIG.png")
    si = eigSileSiesta(sisl_files("siesta", "Si_pdos_k", "Si_pdos.EIG"))
    p, ns = si.ArgumentParser()
    p.parse_args([], namespace=ns)
    p.parse_args(["--energy", " -2:2"], namespace=ns)
    p.parse_args(["--energy", " -2:2", "--plot", png], namespace=ns)


def test_si_pdos_kgrid_eig_ArgumentParser_de(sisl_files, sisl_tmp):
    dat = sisl_tmp("si_pdos_kgrid.EIG.dat")
    kp = sisl_files("siesta", "Si_pdos_k", "Si_pdos.KP")
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.EIG"))
    p, ns = si.ArgumentParser()
    assert ns._dos_args[0] == pytest.approx(0.005)
    assert ns._dos_args[2] == "gaussian"
    assert ns._weights is not None
    ns._weights = None
    p.parse_args(["-kp", str(kp)], namespace=ns)
    assert ns._weights is not None
    p.parse_args(["--dos", "dE=1meV", "kT=20meV"], namespace=ns)
    assert ns._weights is not None
    assert ns._dos_args[0] == pytest.approx(1e-3)
    assert ns._dos_args[1] == pytest.approx(20e-3)
    assert ns._dos_args[2] == "gaussian"
    p.parse_args(["--dos", "kT=10meV", "dist=lorentzian"], namespace=ns)
    assert ns._dos_args[0] == pytest.approx(1e-3)
    assert ns._dos_args[1] == pytest.approx(10e-3)
    assert ns._dos_args[2] == "lorentzian"
    assert len(ns._data) == 3
    p.parse_args(["--dos", "1meV", "5meV", "gaussian"], namespace=ns)
    assert ns._dos_args[0] == pytest.approx(1e-3)
    assert ns._dos_args[1] == pytest.approx(5e-3)
    assert ns._dos_args[2] == "gaussian"
    assert len(ns._data) == 4
    p.parse_args(["--out", dat], namespace=ns)


def test_soc_pt2_xx_eig(sisl_files):
    f = sisl_files("siesta", "Pt2_soc", "Pt2_xx.EIG")
    eig = eigSileSiesta(f).read_data()

    # nspin, nk, nb
    # Since SO/NC mixes spin-channels it makes no sense
    # to have them separately
    assert np.all(eig.shape == (1, 1, 76))


def test_soc_pt2_xx_eig_fermi_level(sisl_files):
    f = sisl_files("siesta", "Pt2_soc", "Pt2_xx.EIG")
    ef = eigSileSiesta(f).read_fermi_level()
    fdf = sisl_files("siesta", "Pt2_soc", "Pt2.fdf")
    ef1 = fdfSileSiesta(fdf).read_fermi_level(order="EIG")
    assert ef == pytest.approx(ef1)
    # This should prefer the TSHS
    ef2 = fdfSileSiesta(fdf).read_fermi_level(order="TSHS")

    # This test is for 5, with codata changes.
    assert ef == pytest.approx(ef2, abs=1e-5)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import operator as op

import numpy as np
import pytest

from sisl.mixing import AndersonMixer, LinearMixer

pytestmark = pytest.mark.mixing


def test_linear_mixer():
    def scf(f):
        return np.cos(f)

    f = np.linspace(0, 7, 1000)
    mix = LinearMixer()
    s = str(mix)

    dmax = 1
    i = 0
    while dmax > 1e-7:
        i += 1
        df = scf(f) - f
        dmax = np.fabs(df).max()
        f = mix(f, df)


def test_anderson_mixer():
    def scf(f):
        return np.cos(f)

    f = np.linspace(0, 7, 1000)
    mix = AndersonMixer()
    s = str(mix)

    dmax = 1
    i = 0
    while dmax > 1e-7:
        i += 1
        df = scf(f) - f
        dmax = np.fabs(df).max()
        f = mix(f, df)


def test_composite_mixer():
    def scf(f):
        return np.cos(f)

    f = np.linspace(0, 7, 1000)
    mix = AndersonMixer() * 0.1 + 0.9 * LinearMixer()
    s = str(mix)

    dmax = 1
    i = 0
    while dmax > 1e-7:
        i += 1
        df = scf(f) - f
        dmax = np.fabs(df).max()
        f = mix(f, df)


@pytest.mark.parametrize("op", [op.add, op.sub, op.mul, op.truediv, op.pow])
def test_composite_mixer_init(op):
    mix1 = AndersonMixer()
    mix2 = LinearMixer()

    f = np.linspace(0, 7, 1000)
    df = f * 0.1

    _ = op(mix1, mix2)(f, df)
    _ = op(mix1, 1)(f, df)
    _ = op(1, mix1)(f, df)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import Atom, Cuboid, Geometry, PeriodicTable
from sisl.geom import *

pytestmark = [pytest.mark.geom, pytest.mark.category, pytest.mark.geom_category]


def test_geom_category_kw_and_call():
    # Check that categories can be built indistinctively using the kw builder
    # or directly calling them.
    cat1 = AtomCategory.kw(odd={})
    cat2 = AtomCategory(odd={})
    assert cat1 == cat2


def test_geom_category():
    hBN = honeycomb(1.42, [Atom(5, R=1.43), Atom(7, R=1.43)]) * (10, 11, 1)

    B = AtomZ([5, 6])
    B2 = AtomNeighbors(2, neighbor=B)
    N = AtomZ(7)
    N2 = AtomNeighbors(2, neighbor=N)
    B3 = AtomNeighbors(3, neighbor=B)
    N3 = AtomNeighbors(3, neighbor=N)

    assert B != N

    n2 = AtomNeighbors(2)
    n3 = AtomNeighbors(3)
    n3n2 = AtomNeighbors(1, 3, neighbor=n2) & n3
    n3n3n2 = AtomNeighbors(1, 3, neighbor=n3n2) & n3

    category = (B & B2) ^ (N & N2) ^ (B & B3) ^ (N & N3) ^ n2

    cat = category.categorize(hBN)
    cat1 = category.categorize(hBN, atoms=[2, 3])
    assert len(cat1) == 2
    assert cat[2] == cat1[0]
    assert cat[3] == cat1[1]


def test_geom_category_no_r():
    hBN = honeycomb(1.42, Atom[5, 7]) * (10, 11, 1)

    B = AtomZ(5)
    B2 = AtomNeighbors(2, neighbor=B, R=1.43)
    N = AtomZ(7)
    N2 = AtomNeighbors(min=2, max=2, neighbor=N, R=(0.01, 1.43))
    PT = PeriodicTable()
    B3 = AtomNeighbors(3, neighbor=B, R=lambda atom: (0.01, PT.radius(atom.Z)))
    N3 = AtomNeighbors(3, neighbor=N, R=1.43)
    Nabove3 = AtomNeighbors(min=3, neighbor=N, R=1.43)

    assert B != N2
    assert N2 != N3

    n2 = AtomNeighbors(2, R=1.43)

    category = (B & B2) ^ (N & N2) ^ (B & B3) ^ (N & N3) ^ n2

    cat = category.categorize(hBN)


def test_geom_category_even_odd():
    hBN = honeycomb(1.42, Atom[5, 7]) * (4, 5, 1)

    odd = AtomOdd()
    even = AtomEven()

    assert odd != even
    assert even != odd

    cat = (odd | even).categorize(hBN)
    for i, c in enumerate(cat):
        if i % 2 == 0:
            assert c == even
        else:
            assert c == odd


def test_geom_category_index():
    hBN = honeycomb(1.42, Atom[5, 7]) * (4, 5, 1)

    first_and_third = AtomIndex([0, 2])
    assert set([0, 2]) == set(hBN.asc2uc(first_and_third))

    odd = AtomIndex(mod=2)
    even = ~odd
    str(odd)
    str(even)

    assert odd != even
    assert even != odd

    cat = (odd | even).categorize(hBN)
    for i, c in enumerate(cat):
        if i % 2 == 0:
            assert c == even
        else:
            assert c == odd


def test_geom_category_seq():
    geom = honeycomb(1.42, Atom[5, 7]) * (4, 5, 1)

    cat = AtomSeq("")
    assert len(geom.asc2uc(cat)) == 0

    cat = AtomSeq("0:")
    assert len(geom.asc2uc(cat)) == geom.na

    cat = AtomSeq(":")
    assert len(geom.asc2uc(cat)) == geom.na

    cat = AtomSeq(":-2")
    assert len(geom.asc2uc(cat)) == geom.na - 1

    cat = AtomSeq("-2:")
    assert len(geom.asc2uc(cat)) == 2

    cat = AtomSeq("-2:2:")
    assert len(geom.asc2uc(cat)) == 1

    cat = AtomSeq("-2")
    assert set(geom.asc2uc(cat)) == set([len(geom) - 1 - 2])

    cat = AtomSeq("0,3,5")
    assert set(geom.asc2uc(cat)) == set([0, 3, 5])

    cat = AtomSeq(":3,5")
    assert set(geom.asc2uc(cat)) == set([0, 1, 2, 3, 5])


def test_geom_category_tag():
    atoms = [
        Atom(Z=6, tag="C1"),
        Atom(Z=6, tag="C2"),
        Atom(Z=6, tag="C3"),
        Atom(Z=1, tag="H"),
    ]
    geom = Geometry([[0, 0, 0]] * 4, atoms=atoms)

    cat = AtomTag("")
    assert len(geom.asc2uc(cat)) == 4

    cat = AtomTag("C")
    assert set(geom.asc2uc(cat)) == set([0, 1, 2])

    cat = AtomTag("C1")
    assert set(geom.asc2uc(cat)) == set([0])

    cat = AtomTag("C$")
    assert len(geom.asc2uc(cat)) == 0

    cat = AtomTag("C[13]")
    assert set(geom.asc2uc(cat)) == set([0, 2])

    cat = AtomTag("[CH]")
    assert set(geom.asc2uc(cat)) == set([0, 1, 2, 3])

    cat = AtomTag("[CH]$")
    assert set(geom.asc2uc(cat)) == set([3])


def test_geom_category_frac_site():
    hBN_gr = bilayer(1.42, Atom[5, 7], Atom[6]) * (4, 5, 1)
    mid_layer = np.average(hBN_gr.xyz[:, 2])

    A_site = AtomFracSite(graphene())
    bottom = AtomXYZ(z_lt=mid_layer, z=(None, mid_layer))
    top = ~AtomCategory(xyz={"z_lt": mid_layer}, z=(None, mid_layer))

    bottom_A = A_site & bottom
    top_A = A_site & top

    cat = (bottom_A | top_A).categorize(hBN_gr)
    str(cat)
    cnull = 0
    for i, c in enumerate(cat):
        if c == NullCategory():
            cnull += 1
        elif hBN_gr.xyz[i, 2] < mid_layer:
            assert c == bottom_A
        else:
            assert False
    assert cnull == len(hBN_gr) // 4 * 3


def test_geom_category_frac_A_B_site():
    gr = graphene() * (4, 5, 1)

    A_site = AtomFracSite(graphene())
    B_site = AtomFracSite(graphene(), foffset=(-1 / 3, -1 / 3, 0))

    cat = (A_site | B_site).categorize(gr)
    for i, c in enumerate(cat):
        if i % 2 == 0:
            assert c == A_site
        else:
            assert c == B_site


def test_geom_category_shape():
    hBN_gr = bilayer(1.42, Atom[5, 7], Atom[6]) * (4, 5, 1)
    mid_layer = np.average(hBN_gr.xyz[:, 2])

    A_site = AtomFracSite(graphene())
    bottom = AtomXYZ(Cuboid([10000, 10000, mid_layer], [-100, -100, 0]))
    top = ~bottom

    bottom_A = A_site & bottom
    top_A = A_site & top

    cat = (bottom_A | top_A).categorize(hBN_gr)
    cnull = 0
    for i, c in enumerate(cat):
        if c == NullCategory():
            cnull += 1
        elif hBN_gr.xyz[i, 2] < mid_layer:
            assert c == bottom_A
        else:
            assert False
    assert cnull == len(hBN_gr) // 4 * 3


def test_geom_category_xyz_none():
    hBN_gr = bilayer(1.42, Atom[5, 7], Atom[6]) * (4, 5, 1)
    mid_layer = np.average(hBN_gr.xyz[:, 2])

    A_site = AtomFracSite(graphene())
    bottom = AtomXYZ(z=(None, mid_layer))
    top = AtomXYZ(z=(mid_layer, None))

    bottom_A = A_site & bottom
    top_A = A_site & top

    cat = (bottom_A | top_A).categorize(hBN_gr)
    cnull = 0
    for i, c in enumerate(cat):
        if c == NullCategory():
            cnull += 1
        elif hBN_gr.xyz[i, 2] < mid_layer:
            assert c == bottom_A
        else:
            assert False
    assert cnull == len(hBN_gr) // 4 * 3


def test_geom_category_xyz_meta():
    """
    We check that the metaclass defined for individual direction categories works.
    """
    hBN_gr = bilayer(1.42, Atom[5, 7], Atom[6]) * (4, 5, 1)
    sc2uc = hBN_gr.asc2uc

    # Check that all classes work
    for key in ("x", "y", "z", "f_x", "f_y", "f_z", "a_x", "a_y", "a_z"):
        name = key.replace("_", "")

        # Check that the attribute is present
        assert hasattr(AtomXYZ, name)

        # Get the category class
        cls = getattr(AtomXYZ, name)

        # Assert that using the class actually does the same effect as calling the AtomXYZ
        # category with the appropiate arguments
        def get_cls(op, v):
            return AtomXYZ(**{f"{key}_{op}": v})

        assert np.all(sc2uc(cls < 0.5) == sc2uc(cls(lt=0.5)))
        assert np.all(sc2uc(cls < 0.5) == sc2uc(get_cls("lt", 0.5)))
        assert np.all(sc2uc(cls <= 0.5) == sc2uc(cls(le=0.5)))
        assert np.all(sc2uc(cls <= 0.5) == sc2uc(get_cls("le", 0.5)))
        assert np.all(sc2uc(cls > 0.5) == sc2uc(cls(gt=0.5)))
        assert np.all(sc2uc(cls > 0.5) == sc2uc(get_cls("gt", 0.5)))
        assert np.all(sc2uc(cls >= 0.5) == sc2uc(cls(ge=0.5)))
        assert np.all(sc2uc(cls >= 0.5) == sc2uc(get_cls("ge", 0.5)))

        assert np.all(
            sc2uc(cls((-1, 1))) == sc2uc(get_cls("ge", -1) & get_cls("le", 1))
        )
        assert np.all(sc2uc(cls(-1, 1)) == sc2uc(get_cls("ge", -1) & get_cls("le", 1)))
